@@ -1,6 +1,6 @@
 /*
  * ProFTPD - mod_sftp MACs
- * Copyright (c) 2008-2009 TJ Saunders
+ * Copyright (c) 2008-2015 TJ Saunders
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,14 +14,12 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, TJ Saunders and other respective copyright holders
  * give permission to link this program with OpenSSL, and distribute the
  * resulting executable, without including the source code for OpenSSL in the
  * source distribution.
- *
- * $Id: mac.c,v 1.5 2009/09/16 20:51:15 castaglia Exp $
  */
 
 #include "mod_sftp.h"
@@ -33,9 +31,12 @@
 #include "session.h"
 #include "disconnect.h"
 #include "interop.h"
+#include "umac.h"
 
 struct sftp_mac {
   const char *algo;
+  int algo_type;
+
   const EVP_MD *digest;
 
   unsigned char *key;
@@ -49,6 +50,12 @@ struct sftp_mac {
   uint32_t mac_len;
 };
 
+#define SFTP_MAC_ALGO_TYPE_HMAC		1
+#define SFTP_MAC_ALGO_TYPE_UMAC64	2
+
+#define SFTP_MAC_FL_READ_MAC	1
+#define SFTP_MAC_FL_WRITE_MAC	2
+
 /* We need to keep the old MACs around, so that we can handle N arbitrary
  * packets to/from the client using the old keys, as during rekeying.
  * Thus we have two read MAC contexts, two write MAC contexts.
@@ -56,19 +63,24 @@ struct sftp_mac {
  */
 
 static struct sftp_mac read_macs[] = {
-  { NULL, NULL, NULL, 0 },
-  { NULL, NULL, NULL, 0 }
+  { NULL, 0, NULL, NULL, 0 },
+  { NULL, 0, NULL, NULL, 0 }
 };
-static HMAC_CTX read_ctxs[2];
+static HMAC_CTX hmac_read_ctxs[2];
+static struct umac_ctx *umac_read_ctxs[2];
 
 static struct sftp_mac write_macs[] = {
-  { NULL, NULL, NULL, 0 },
-  { NULL, NULL, NULL, 0 }
+  { NULL, 0, NULL, NULL, 0 },
+  { NULL, 0, NULL, NULL, 0 }
 };
-static HMAC_CTX write_ctxs[2];
+static HMAC_CTX hmac_write_ctxs[2];
+static struct umac_ctx *umac_write_ctxs[2];
 
 static size_t mac_blockszs[2] = { 0, 0 };
 
+/* Buffer size for reading/writing keys */
+#define SFTP_MAC_BUFSZ				4096
+
 static unsigned int read_mac_idx = 0;
 static unsigned int write_mac_idx = 0;
 
@@ -93,10 +105,12 @@ static void switch_read_mac(void) {
   if (read_macs[read_mac_idx].key) {
     clear_mac(&(read_macs[read_mac_idx]));
 #if OPENSSL_VERSION_NUMBER > 0x000907000L
-    HMAC_CTX_cleanup(&(read_ctxs[read_mac_idx]));
+    HMAC_CTX_cleanup(&(hmac_read_ctxs[read_mac_idx]));
 #else
-    HMAC_cleanup(&(read_ctxs[read_mac_idx]));
+    HMAC_cleanup(&(hmac_read_ctxs[read_mac_idx]));
 #endif
+    umac_reset(umac_read_ctxs[read_mac_idx]);
+
     mac_blockszs[read_mac_idx] = 0; 
 
     /* Now we can switch the index. */
@@ -114,10 +128,11 @@ static void switch_write_mac(void) {
   if (write_macs[write_mac_idx].key) {
     clear_mac(&(write_macs[write_mac_idx]));
 #if OPENSSL_VERSION_NUMBER > 0x000907000L
-    HMAC_CTX_cleanup(&(write_ctxs[write_mac_idx]));
+    HMAC_CTX_cleanup(&(hmac_write_ctxs[write_mac_idx]));
 #else
-    HMAC_cleanup(&(write_ctxs[write_mac_idx]));
+    HMAC_cleanup(&(hmac_write_ctxs[write_mac_idx]));
 #endif
+    umac_reset(umac_write_ctxs[write_mac_idx]);
 
     /* Now we can switch the index. */
     if (write_mac_idx == 1) {
@@ -142,21 +157,229 @@ static void clear_mac(struct sftp_mac *mac) {
   mac->algo = NULL;
 }
 
+static int init_mac(pool *p, struct sftp_mac *mac, HMAC_CTX *hmac_ctx,
+    struct umac_ctx *umac_ctx) {
+#if OPENSSL_VERSION_NUMBER > 0x000907000L
+  HMAC_CTX_init(hmac_ctx);
+#else
+  /* Reset the HMAC context. */
+  HMAC_Init(hmac_ctx, NULL, 0, NULL);
+#endif
+  umac_reset(umac_ctx);
+
+  if (mac->algo_type == SFTP_MAC_ALGO_TYPE_HMAC) {
+#if OPENSSL_VERSION_NUMBER > 0x000907000L
+# if OPENSSL_VERSION_NUMBER >= 0x10000001L
+    if (HMAC_Init_ex(hmac_ctx, mac->key, mac->key_len, mac->digest,
+        NULL) != 1) {
+      pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+        "error initializing HMAC: %s", sftp_crypto_get_errors());
+      errno = EPERM;
+      return -1;
+    }
+
+# else
+    HMAC_Init_ex(hmac_ctx, mac->key, mac->key_len, mac->digest, NULL);
+# endif /* OpenSSL-1.0.0 and later */
+
+#else
+    HMAC_Init(hmac_ctx, mac->key, mac->key_len, mac->digest);
+#endif
+
+  } else if (mac->algo_type == SFTP_MAC_ALGO_TYPE_UMAC64) {
+    umac_init(umac_ctx, mac->key);
+  }
+
+  return 0;
+}
+
+static int get_mac(struct ssh2_packet *pkt, struct sftp_mac *mac,
+    HMAC_CTX *hmac_ctx, struct umac_ctx *umac_ctx, int flags) {
+  unsigned char *mac_data;
+  unsigned char *buf, *ptr;
+  uint32_t buflen, bufsz = 0, mac_len = 0;
+
+  if (mac->algo_type == SFTP_MAC_ALGO_TYPE_HMAC) {
+    bufsz = (sizeof(uint32_t) * 2) + pkt->packet_len;
+    mac_data = pcalloc(pkt->pool, EVP_MAX_MD_SIZE);
+
+    buflen = bufsz;
+    ptr = buf = sftp_msg_getbuf(pkt->pool, bufsz);
+
+    sftp_msg_write_int(&buf, &buflen, pkt->seqno);
+    sftp_msg_write_int(&buf, &buflen, pkt->packet_len);
+    sftp_msg_write_byte(&buf, &buflen, pkt->padding_len);
+    sftp_msg_write_data(&buf, &buflen, pkt->payload, pkt->payload_len, FALSE);
+    sftp_msg_write_data(&buf, &buflen, pkt->padding, pkt->padding_len, FALSE);
+
+#if OPENSSL_VERSION_NUMBER > 0x000907000L
+# if OPENSSL_VERSION_NUMBER >= 0x10000001L
+    if (HMAC_Init_ex(hmac_ctx, NULL, 0, NULL, NULL) != 1) {
+      pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+        "error resetting HMAC context: %s", sftp_crypto_get_errors());
+      errno = EPERM;
+      return -1;
+    }
+# else
+    HMAC_Init_ex(hmac_ctx, NULL, 0, NULL, NULL);
+# endif /* OpenSSL-1.0.0 and later */
+
+#else
+    HMAC_Init(hmac_ctx, NULL, 0, NULL);
+#endif /* OpenSSL-0.9.7 and later */
+
+#if OPENSSL_VERSION_NUMBER >= 0x10000001L
+    if (HMAC_Update(hmac_ctx, ptr, (bufsz - buflen)) != 1) {
+      pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+        "error adding %lu bytes of data to  HMAC context: %s",
+        (unsigned long) (bufsz - buflen), sftp_crypto_get_errors());
+      errno = EPERM;
+      return -1;
+    }
+
+    if (HMAC_Final(hmac_ctx, mac_data, &mac_len) != 1) {
+      pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+        "error finalizing HMAC context: %s", sftp_crypto_get_errors());
+      errno = EPERM;
+      return -1;
+    }
+#else
+    HMAC_Update(hmac_ctx, ptr, (bufsz - buflen));
+    HMAC_Final(hmac_ctx, mac_data, &mac_len);
+#endif /* OpenSSL-1.0.0 and later */
+
+  } else if (mac->algo_type == SFTP_MAC_ALGO_TYPE_UMAC64) {
+    unsigned char nonce[8], *nonce_ptr;
+    uint32_t nonce_len = 0;
+
+    bufsz = sizeof(uint32_t) + pkt->packet_len;
+    mac_data = pcalloc(pkt->pool, EVP_MAX_MD_SIZE);
+
+    buflen = bufsz;
+    ptr = buf = sftp_msg_getbuf(pkt->pool, bufsz);
+
+    sftp_msg_write_int(&buf, &buflen, pkt->packet_len);
+    sftp_msg_write_byte(&buf, &buflen, pkt->padding_len);
+    sftp_msg_write_data(&buf, &buflen, pkt->payload, pkt->payload_len, FALSE);
+    sftp_msg_write_data(&buf, &buflen, pkt->padding, pkt->padding_len, FALSE);
+
+    nonce_ptr = nonce;
+    nonce_len = sizeof(nonce);
+    sftp_msg_write_long(&nonce_ptr, &nonce_len, pkt->seqno);
+
+    umac_reset(umac_ctx);
+    umac_update(umac_ctx, ptr, (bufsz - buflen));
+    umac_final(umac_ctx, mac_data, nonce);
+    mac_len = 8;
+  }
+
+  if (mac_len == 0) {
+    pkt->mac = NULL;
+    pkt->mac_len = 0;
+
+    (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+      "error computing MAC using %s: %s", mac->algo,
+      sftp_crypto_get_errors());
+
+    errno = EIO;
+    return -1;
+  }
+
+  if (mac->mac_len != 0) {
+    mac_len = mac->mac_len;
+  }
+
+  if (flags & SFTP_MAC_FL_READ_MAC) {
+    if (memcmp(mac_data, pkt->mac, mac_len) != 0) {
+      unsigned int i = 0;
+
+      (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+        "MAC from client differs from expected MAC using %s", mac->algo);
+
+#ifdef SFTP_DEBUG_PACKET
+      (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+        "client MAC (len %lu):", (unsigned long) pkt->mac_len);
+      for (i = 0; i < mac_len;) {
+        (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+          "  %02x%02x %02x%02x %02x%02x %02x%02x",
+          ((unsigned char *) pkt->mac)[i], ((unsigned char *) pkt->mac)[i+1],
+          ((unsigned char *) pkt->mac)[i+2], ((unsigned char *) pkt->mac)[i+3],
+          ((unsigned char *) pkt->mac)[i+4], ((unsigned char *) pkt->mac)[i+5],
+          ((unsigned char *) pkt->mac)[i+6], ((unsigned char *) pkt->mac)[i+7]);
+        i += 8;
+      }
+
+      (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+        "server MAC (len %lu):", (unsigned long) mac_len);
+      for (i = 0; i < mac_len;) {
+        (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+          "  %02x%02x %02x%02x %02x%02x %02x%02x",
+          ((unsigned char *) mac)[i], ((unsigned char *) mac)[i+1],
+          ((unsigned char *) mac)[i+2], ((unsigned char *) mac)[i+3],
+          ((unsigned char *) mac)[i+4], ((unsigned char *) mac)[i+5],
+          ((unsigned char *) mac)[i+6], ((unsigned char *) mac)[i+7]);
+        i += 8;
+      }
+#else
+      /* Avoid compiler warning. */
+      (void) i;
+#endif
+
+      errno = EINVAL;
+      return -1;
+    }
+
+  } else if (flags & SFTP_MAC_FL_WRITE_MAC) {
+    unsigned int i = 0;
+#ifdef SFTP_DEBUG_PACKET
+
+    if (pkt->mac_len > 0) {
+      (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+        "server MAC (len %lu, seqno %lu):",
+        (unsigned long) pkt->mac_len, (unsigned long) pkt->seqno);
+      for (i = 0; i < mac_len;) {
+        (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+          "  %02x%02x %02x%02x %02x%02x %02x%02x",
+          ((unsigned char *) pkt->mac)[i], ((unsigned char *) pkt->mac)[i+1],
+          ((unsigned char *) pkt->mac)[i+2], ((unsigned char *) pkt->mac)[i+3],
+          ((unsigned char *) pkt->mac)[i+4], ((unsigned char *) pkt->mac)[i+5],
+          ((unsigned char *) pkt->mac)[i+6], ((unsigned char *) pkt->mac)[i+7]);
+        i += 8;
+      }
+    }
+#else
+    /* Avoid compiler warning. */
+    (void) i;
+#endif
+  }
+
+  pkt->mac_len = mac_len;
+  pkt->mac = pcalloc(pkt->pool, pkt->mac_len);
+  memcpy(pkt->mac, mac_data, mac_len);
+
+  return 0;
+}
+
 static int set_mac_key(struct sftp_mac *mac, const EVP_MD *hash,
-    const char *k, uint32_t klen, const char *h, uint32_t hlen, char *letter,
-    const unsigned char *id, uint32_t id_len) {
- 
+    const unsigned char *k, uint32_t klen, const char *h, uint32_t hlen,
+    char *letter, const unsigned char *id, uint32_t id_len) {
   EVP_MD_CTX ctx;
   unsigned char *key = NULL;
   size_t key_sz;
   uint32_t key_len = 0;
- 
+
   key_sz = sftp_crypto_get_size(EVP_MD_block_size(mac->digest),
     EVP_MD_size(hash)); 
+  if (key_sz == 0) {
+    (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+      "unable to determine key length for MAC '%s'", mac->algo);
+    errno = EINVAL;
+    return -1;
+  }
 
   key = malloc(key_sz);
   if (key == NULL) {
-    pr_log_pri(PR_LOG_CRIT, MOD_SFTP_VERSION ": Out of memory!");
+    pr_log_pri(PR_LOG_ALERT, MOD_SFTP_VERSION ": Out of memory!");
     _exit(1);
   }
 
@@ -169,6 +392,7 @@ static int set_mac_key(struct sftp_mac *mac, const EVP_MD *hash,
   if (EVP_DigestInit(&ctx, hash) != 1) {
     (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
       "error initializing message digest: %s", sftp_crypto_get_errors());
+    free(key);
     return -1;
   }
 #else
@@ -179,6 +403,7 @@ static int set_mac_key(struct sftp_mac *mac, const EVP_MD *hash,
   if (EVP_DigestUpdate(&ctx, k, klen) != 1) {
     (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
       "error updating message digest with K: %s", sftp_crypto_get_errors());
+    free(key);
     return -1;
   }
 #else
@@ -189,6 +414,7 @@ static int set_mac_key(struct sftp_mac *mac, const EVP_MD *hash,
   if (EVP_DigestUpdate(&ctx, h, hlen) != 1) {
     (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
       "error updating message digest with H: %s", sftp_crypto_get_errors());
+    free(key);
     return -1;
   }
 #else
@@ -200,6 +426,7 @@ static int set_mac_key(struct sftp_mac *mac, const EVP_MD *hash,
     (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
       "error updating message digest with '%c': %s", *letter,
       sftp_crypto_get_errors());
+    free(key);
     return -1;
   }
 #else
@@ -210,6 +437,7 @@ static int set_mac_key(struct sftp_mac *mac, const EVP_MD *hash,
   if (EVP_DigestUpdate(&ctx, (char *) id, id_len) != 1) {
     (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
       "error updating message digest with ID: %s", sftp_crypto_get_errors());
+    free(key);
     return -1;
   }
 #else
@@ -220,6 +448,8 @@ static int set_mac_key(struct sftp_mac *mac, const EVP_MD *hash,
   if (EVP_DigestFinal(&ctx, key, &key_len) != 1) {
     (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
       "error finalizing message digest: %s", sftp_crypto_get_errors());
+    pr_memscrub(key, key_sz);
+    free(key);
     return -1;
   }
 #else
@@ -229,6 +459,7 @@ static int set_mac_key(struct sftp_mac *mac, const EVP_MD *hash,
   /* If we need more, keep hashing, as per RFC, until we have enough
    * material.
    */
+
   while (key_sz > key_len) {
     uint32_t len = key_len;
 
@@ -238,6 +469,8 @@ static int set_mac_key(struct sftp_mac *mac, const EVP_MD *hash,
     if (EVP_DigestInit(&ctx, hash) != 1) {
       (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
         "error initializing message digest: %s", sftp_crypto_get_errors());
+      pr_memscrub(key, key_sz);
+      free(key);
       return -1;
     }
 #else
@@ -248,6 +481,8 @@ static int set_mac_key(struct sftp_mac *mac, const EVP_MD *hash,
     if (EVP_DigestUpdate(&ctx, k, klen) != 1) {
       (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
         "error updating message digest with K: %s", sftp_crypto_get_errors());
+      pr_memscrub(key, key_sz);
+      free(key);
       return -1;
     }
 #else
@@ -258,6 +493,8 @@ static int set_mac_key(struct sftp_mac *mac, const EVP_MD *hash,
     if (EVP_DigestUpdate(&ctx, h, hlen) != 1) {
       (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
         "error updating message digest with H: %s", sftp_crypto_get_errors());
+      pr_memscrub(key, key_sz);
+      free(key);
       return -1;
     }
 #else
@@ -269,6 +506,8 @@ static int set_mac_key(struct sftp_mac *mac, const EVP_MD *hash,
       (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
         "error updating message digest with data: %s",
         sftp_crypto_get_errors());
+      pr_memscrub(key, key_sz);
+      free(key);
       return -1;
     }
 #else
@@ -279,6 +518,8 @@ static int set_mac_key(struct sftp_mac *mac, const EVP_MD *hash,
     if (EVP_DigestFinal(&ctx, key + len, &len) != 1) {
       (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
         "error finalizing message digest: %s", sftp_crypto_get_errors());
+      pr_memscrub(key, key_sz);
+      free(key);
       return -1;
     }
 #else
@@ -291,7 +532,13 @@ static int set_mac_key(struct sftp_mac *mac, const EVP_MD *hash,
   mac->key = key;
   mac->keysz = key_sz;
 
-  mac->key_len = EVP_MD_size(mac->digest);
+  if (mac->algo_type == SFTP_MAC_ALGO_TYPE_HMAC) {
+    mac->key_len = EVP_MD_size(mac->digest);
+
+  } else if (mac->algo_type == SFTP_MAC_ALGO_TYPE_UMAC64) {
+    mac->key_len = EVP_MD_block_size(mac->digest);
+  }
+
   if (!sftp_interop_supports_feature(SFTP_SSH2_FEAT_MAC_LEN)) {
     mac->key_len = 16;
   }
@@ -319,7 +566,7 @@ const char *sftp_mac_get_read_algo(void) {
 
 int sftp_mac_set_read_algo(const char *algo) {
   uint32_t mac_len;
-  unsigned int idx = write_mac_idx;
+  unsigned int idx = read_mac_idx;
 
   if (read_macs[idx].key) {
     /* If we have an existing key, it means that we are currently rekeying. */
@@ -327,10 +574,18 @@ int sftp_mac_set_read_algo(const char *algo) {
   }
 
   read_macs[idx].digest = sftp_crypto_get_digest(algo, &mac_len);
-  if (read_macs[idx].digest == NULL)
+  if (read_macs[idx].digest == NULL) {
     return -1;
+  }
 
   read_macs[idx].algo = algo;
+  if (strncmp(read_macs[idx].algo, "umac-64@openssh.com", 12) == 0) {
+    read_macs[idx].algo_type = SFTP_MAC_ALGO_TYPE_UMAC64;
+
+  } else {
+    read_macs[idx].algo_type = SFTP_MAC_ALGO_TYPE_HMAC;
+  }
+
   read_macs[idx].mac_len = mac_len;
   return 0;
 }
@@ -338,19 +593,21 @@ int sftp_mac_set_read_algo(const char *algo) {
 int sftp_mac_set_read_key(pool *p, const EVP_MD *hash, const BIGNUM *k,
     const char *h, uint32_t hlen) {
   const unsigned char *id = NULL;
-  char *buf, *ptr;
+  unsigned char *buf, *ptr;
   uint32_t buflen, bufsz, id_len;
   char letter;
   size_t blocksz;
   struct sftp_mac *mac;
-  HMAC_CTX *mac_ctx;
+  HMAC_CTX *hmac_ctx;
+  struct umac_ctx *umac_ctx;
 
   switch_read_mac();
 
   mac = &(read_macs[read_mac_idx]);
-  mac_ctx = &(read_ctxs[read_mac_idx]);
+  hmac_ctx = &(hmac_read_ctxs[read_mac_idx]);
+  umac_ctx = umac_read_ctxs[read_mac_idx];
 
-  bufsz = buflen = 1024;
+  bufsz = buflen = SFTP_MAC_BUFSZ;
   ptr = buf = sftp_msg_getbuf(p, bufsz);
 
   /* Need to use SSH2-style format of K for the key. */
@@ -362,13 +619,9 @@ int sftp_mac_set_read_key(pool *p, const EVP_MD *hash, const BIGNUM *k,
   letter = 'E';
   set_mac_key(mac, hash, ptr, (bufsz - buflen), h, hlen, &letter, id, id_len);
 
-#if OPENSSL_VERSION_NUMBER > 0x000907000L
-  HMAC_CTX_init(mac_ctx);
-#else
-  /* Reset the HMAC context. */
-  HMAC_Init(mac_ctx, NULL, 0, NULL);
-#endif
-  HMAC_Init(mac_ctx, mac->key, mac->key_len, mac->digest);
+  if (init_mac(p, mac, hmac_ctx, umac_ctx) < 0) {
+    return -1;
+  }
 
   if (mac->mac_len == 0) {
     blocksz = EVP_MD_size(mac->digest);
@@ -384,89 +637,26 @@ int sftp_mac_set_read_key(pool *p, const EVP_MD *hash, const BIGNUM *k,
 
 int sftp_mac_read_data(struct ssh2_packet *pkt) {
   struct sftp_mac *mac;
-  HMAC_CTX *mac_ctx;
+  HMAC_CTX *hmac_ctx;
+  struct umac_ctx *umac_ctx;
+  int res;
 
   mac = &(read_macs[read_mac_idx]);
-  mac_ctx = &(read_ctxs[read_mac_idx]);
+  hmac_ctx = &(hmac_read_ctxs[read_mac_idx]);
+  umac_ctx = umac_read_ctxs[read_mac_idx];
 
-  if (mac->key) {
-    unsigned char *mac_data;
-    char *buf, *ptr;
-    uint32_t buflen, bufsz = (sizeof(uint32_t) * 2) + pkt->packet_len,
-      mac_len = 0;
-
-    mac_data = pcalloc(pkt->pool, EVP_MAX_MD_SIZE);
-
-    buflen = bufsz;
-    ptr = buf = sftp_msg_getbuf(pkt->pool, bufsz);
-
-    sftp_msg_write_int(&buf, &buflen, pkt->seqno);
-    sftp_msg_write_int(&buf, &buflen, pkt->packet_len);
-    sftp_msg_write_byte(&buf, &buflen, pkt->padding_len);
-    sftp_msg_write_data(&buf, &buflen, pkt->payload, pkt->payload_len, FALSE);
-    sftp_msg_write_data(&buf, &buflen, pkt->padding, pkt->padding_len, FALSE);
-
-    HMAC_Init(mac_ctx, NULL, 0, NULL);
-    HMAC_Update(mac_ctx, (unsigned char *) ptr, (bufsz - buflen));
-    HMAC_Final(mac_ctx, mac_data, &mac_len);
-
-    if (mac_len == 0) {
-      (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
-        "error computing MAC using %s: %s", mac->algo,
-        sftp_crypto_get_errors());
-      SFTP_DISCONNECT_CONN(SFTP_SSH2_DISCONNECT_MAC_ERROR, NULL);
-    }
-
-    if (mac->mac_len != 0) {
-      mac_len = mac->mac_len;
-    }
-
-    if (mac_len != pkt->mac_len) {
-      (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
-        "expected %u MAC len from client, got %lu", mac_len,
-        (unsigned long) pkt->mac_len);
-      SFTP_DISCONNECT_CONN(SFTP_SSH2_DISCONNECT_MAC_ERROR, NULL);
-    }
-
-    if (memcmp(mac_data, pkt->mac, mac_len) != 0) {
-      (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
-        "MAC from client differs from expected MAC using %s", mac->algo);
-
-#ifdef SFTP_DEBUG_PACKET
-{
-      unsigned int i;
-      (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
-        "client MAC (len %lu):", (unsigned long) pkt->mac_len);
-      for (i = 0; i < mac_len;) {
-        (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
-          "  %02x%02x %02x%02x %02x%02x %02x%02x",
-          ((unsigned char *) pkt->mac)[i], ((unsigned char *) pkt->mac)[i+1],
-          ((unsigned char *) pkt->mac)[i+2], ((unsigned char *) pkt->mac)[i+3],
-          ((unsigned char *) pkt->mac)[i+4], ((unsigned char *) pkt->mac)[i+5],
-          ((unsigned char *) pkt->mac)[i+6], ((unsigned char *) pkt->mac)[i+7]);
-        i += 8;
-      }
-
-      (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
-        "server MAC (len %lu):", (unsigned long) mac_len);
-      for (i = 0; i < mac_len;) {
-        (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
-          "  %02x%02x %02x%02x %02x%02x %02x%02x",
-          ((unsigned char *) mac)[i], ((unsigned char *) mac)[i+1],
-          ((unsigned char *) mac)[i+2], ((unsigned char *) mac)[i+3],
-          ((unsigned char *) mac)[i+4], ((unsigned char *) mac)[i+5],
-          ((unsigned char *) mac)[i+6], ((unsigned char *) mac)[i+7]);
-        i += 8;
-      }
-}
-#endif
-
-      SFTP_DISCONNECT_CONN(SFTP_SSH2_DISCONNECT_MAC_ERROR, NULL);
-    }
+  if (mac->key == NULL) {
+    pkt->mac = NULL;
+    pkt->mac_len = 0;
 
     return 0;
   }
 
+  res = get_mac(pkt, mac, hmac_ctx, umac_ctx, SFTP_MAC_FL_READ_MAC);
+  if (res < 0) {
+    return -1;
+  }
+
   return 0;
 }
 
@@ -488,10 +678,18 @@ int sftp_mac_set_write_algo(const char *algo) {
   }
 
   write_macs[idx].digest = sftp_crypto_get_digest(algo, &mac_len);
-  if (write_macs[idx].digest == NULL)
+  if (write_macs[idx].digest == NULL) {
     return -1;
+  }
 
   write_macs[idx].algo = algo;
+  if (strncmp(write_macs[idx].algo, "umac-64@openssh.com", 12) == 0) {
+    write_macs[idx].algo_type = SFTP_MAC_ALGO_TYPE_UMAC64;
+
+  } else {
+    write_macs[idx].algo_type = SFTP_MAC_ALGO_TYPE_HMAC;
+  }
+
   write_macs[idx].mac_len = mac_len;
   return 0;
 }
@@ -499,18 +697,20 @@ int sftp_mac_set_write_algo(const char *algo) {
 int sftp_mac_set_write_key(pool *p, const EVP_MD *hash, const BIGNUM *k,
     const char *h, uint32_t hlen) {
   const unsigned char *id = NULL;
-  char *buf, *ptr;
+  unsigned char *buf, *ptr;
   uint32_t buflen, bufsz, id_len;
   char letter;
   struct sftp_mac *mac;
-  HMAC_CTX *mac_ctx;
+  HMAC_CTX *hmac_ctx;
+  struct umac_ctx *umac_ctx;
 
   switch_write_mac();
 
   mac = &(write_macs[write_mac_idx]);
-  mac_ctx = &(write_ctxs[write_mac_idx]);
+  hmac_ctx = &(hmac_write_ctxs[write_mac_idx]);
+  umac_ctx = umac_write_ctxs[write_mac_idx];
 
-  bufsz = buflen = 1024;
+  bufsz = buflen = SFTP_MAC_BUFSZ;
   ptr = buf = sftp_msg_getbuf(p, bufsz);
 
   /* Need to use SSH2-style format of K for the key. */
@@ -522,13 +722,9 @@ int sftp_mac_set_write_key(pool *p, const EVP_MD *hash, const BIGNUM *k,
   letter = 'F';
   set_mac_key(mac, hash, ptr, (bufsz - buflen), h, hlen, &letter, id, id_len);
 
-#if OPENSSL_VERSION_NUMBER > 0x000907000L
-  HMAC_CTX_init(mac_ctx);
-#else
-  /* Reset the HMAC context. */
-  HMAC_Init(mac_ctx, NULL, 0, NULL);
-#endif
-  HMAC_Init(mac_ctx, mac->key, mac->key_len, mac->digest);
+  if (init_mac(p, mac, hmac_ctx, umac_ctx) < 0) {
+    return -1;
+  }
 
   pr_memscrub(ptr, bufsz);
   return 0;
@@ -536,75 +732,53 @@ int sftp_mac_set_write_key(pool *p, const EVP_MD *hash, const BIGNUM *k,
 
 int sftp_mac_write_data(struct ssh2_packet *pkt) {
   struct sftp_mac *mac;
-  HMAC_CTX *mac_ctx;
+  HMAC_CTX *hmac_ctx;
+  struct umac_ctx *umac_ctx;
+  int res;
 
   mac = &(write_macs[write_mac_idx]);
-  mac_ctx = &(write_ctxs[write_mac_idx]);
+  hmac_ctx = &(hmac_write_ctxs[write_mac_idx]);
+  umac_ctx = umac_write_ctxs[write_mac_idx];
 
-  if (mac->key) {
-    unsigned char *mac_data;
-    char *buf, *ptr;
-    uint32_t buflen, bufsz = (sizeof(uint32_t) * 2) + pkt->packet_len,
-      mac_len = 0;
-
-    mac_data = pcalloc(pkt->pool, EVP_MAX_MD_SIZE);
-
-    buflen = bufsz;
-    ptr = buf = sftp_msg_getbuf(pkt->pool, bufsz);
-
-    sftp_msg_write_int(&buf, &buflen, pkt->seqno);
-    sftp_msg_write_int(&buf, &buflen, pkt->packet_len);
-    sftp_msg_write_byte(&buf, &buflen, pkt->padding_len);
-    sftp_msg_write_data(&buf, &buflen, pkt->payload, pkt->payload_len, FALSE);
-    sftp_msg_write_data(&buf, &buflen, pkt->padding, pkt->padding_len, FALSE);
-
-    HMAC_Init(mac_ctx, NULL, 0, NULL);
-    HMAC_Update(mac_ctx, (unsigned char *) ptr, (bufsz - buflen));
-    HMAC_Final(mac_ctx, mac_data, &mac_len);
-
-    if (mac_len == 0) {
-      pkt->mac = NULL;
-      pkt->mac_len = 0;
-
-      (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
-        "error computing MAC using %s: %s", mac->algo,
-        sftp_crypto_get_errors());
-      return -1;
-    }
-
-    if (mac->mac_len != 0) {
-      mac_len = mac->mac_len;
-    }
-
-    pkt->mac_len = mac_len;
-    pkt->mac = pcalloc(pkt->pool, pkt->mac_len);
-    memcpy(pkt->mac, mac_data, mac_len);
-
-#ifdef SFTP_DEBUG_PACKET
-{
-  unsigned int i;
-
-  (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
-    "server MAC (len %lu, seqno %lu):",
-    (unsigned long) pkt->mac_len, (unsigned long) pkt->seqno);
-  for (i = 0; i < mac_len;) {
-    (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
-      "  %02x%02x %02x%02x %02x%02x %02x%02x",
-      ((unsigned char *) pkt->mac)[i], ((unsigned char *) pkt->mac)[i+1],
-      ((unsigned char *) pkt->mac)[i+2], ((unsigned char *) pkt->mac)[i+3],
-      ((unsigned char *) pkt->mac)[i+4], ((unsigned char *) pkt->mac)[i+5],
-      ((unsigned char *) pkt->mac)[i+6], ((unsigned char *) pkt->mac)[i+7]);
-    i += 8;
-  }
-}
-#endif
+  if (mac->key == NULL) {
+    pkt->mac = NULL;
+    pkt->mac_len = 0;
 
     return 0;
   }
 
-  pkt->mac = NULL;
-  pkt->mac_len = 0;
+  res = get_mac(pkt, mac, hmac_ctx, umac_ctx, SFTP_MAC_FL_WRITE_MAC);
+  if (res < 0) {
+    return -1;
+  }
 
   return 0;
 }
 
+int sftp_mac_init(void) {
+
+  umac_read_ctxs[0] = umac_alloc();
+  umac_read_ctxs[1] = umac_alloc();
+
+  umac_write_ctxs[0] = umac_alloc();
+  umac_write_ctxs[1] = umac_alloc();
+
+  return 0;
+}
+
+int sftp_mac_free(void) {
+
+  umac_delete(umac_read_ctxs[0]);
+  umac_read_ctxs[0] = NULL;
+
+  umac_delete(umac_read_ctxs[1]);
+  umac_read_ctxs[1] = NULL;
+
+  umac_delete(umac_write_ctxs[0]);
+  umac_write_ctxs[0] = NULL;
+
+  umac_delete(umac_write_ctxs[1]);
+  umac_write_ctxs[1] = NULL;
+
+  return 0;
+}
