@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2011 The ProFTPD Project team
+ * Copyright (c) 2001-2013 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,7 +25,7 @@
  */
 
 /* Configuration structure, server, command and associated prototypes.
- * $Id: dirtree.h,v 1.81 2011/05/23 20:35:35 castaglia Exp $
+ * $Id: dirtree.h,v 1.88 2013-07-16 19:06:13 castaglia Exp $
  */
 
 #ifndef PR_DIRTREE_H
@@ -39,10 +39,17 @@ typedef struct config_struc config_rec;
 
 struct conn_struc;
 
+struct tcp_keepalive {
+  int keepalive_enabled;
+  int keepalive_idle;
+  int keepalive_count;
+  int keepalive_intvl;
+};
+
 typedef struct server_struc {
   struct server_struc *next, *prev;
 
-  pool *pool;			/* Memory pool for this server */
+  struct pool_rec *pool;	/* Memory pool for this server */
   xaset_t *set;			/* Set holding all servers */
 
   /* The label/name for this server configuration. */
@@ -58,8 +65,11 @@ typedef struct server_struc {
    */
   unsigned int ServerPort;
 
-  /* TCP settings: max segment size, receive/send buffer sizes.
+  /* TCP settings: keepalive, max segment size, receive/send buffer sizes.
    */
+
+  struct tcp_keepalive *tcp_keepalive;
+
   int tcp_mss_len;
 
   /* If the tcp_rcvbuf_override/tcp_sndbuf_override flags are true, then
@@ -93,13 +103,16 @@ typedef struct server_struc {
   /* Internal server ID, automatically assigned */
   unsigned int sid;
 
+  /* Private data for passing among modules for this vhost. */
+  pr_table_t *notes;
+
 } server_rec;
 
 typedef struct cmd_struc {
-  pool *pool;
+  struct pool_rec *pool;
   server_rec *server;
   config_rec *config;
-  pool *tmp_pool;		/* Temporary pool which only exists
+  struct pool_rec *tmp_pool;	/* Temporary pool which only exists
 				 * while the cmd's handler is running
 				 */
   int argc;
@@ -108,11 +121,19 @@ typedef struct cmd_struc {
   char **argv;
   char *group;			/* Command grouping */
 
-  int  class;			/* The command class */
+  int  cmd_class;		/* The command class */
   int  stash_index;		/* hack to speed up symbol hashing in modules.c */
   pr_table_t *notes;		/* Private data for passing/retaining between handlers */
 
   int cmd_id;			/* Index into commands list, for faster comparisons */
+
+  /* If we detect that the client sent commands for a protocol OTHER than
+   * FTP, then this field will be FALSE; the protocol field will identify
+   * the detected protocol.
+   */
+  int is_ftp;
+  const char *protocol;
+
 } cmd_rec;
 
 struct config_struc {
@@ -121,7 +142,7 @@ struct config_struc {
   int config_type;
   unsigned int config_id;
 
-  pool *pool;			/* memory pool for this object */
+  struct pool_rec *pool;	/* Memory pool for this object */
   xaset_t *set;			/* The set we are stored in */
   char *name;
   int argc;
@@ -231,12 +252,29 @@ config_rec *add_config_param_str(const char *, int, ...);
 config_rec *add_config_param_set(xaset_t **, const char *, int, ...);
 config_rec *pr_conf_add_server_config_param_str(server_rec *, const char *,
   int, ...);
+
+/* Flags used when searching for specific config_recs in the in-memory
+ * config database, particularly when 'recurse' is TRUE.
+ */
+#define PR_CONFIG_FIND_FL_SKIP_ANON		0x001
+#define PR_CONFIG_FIND_FL_SKIP_DIR		0x002
+#define PR_CONFIG_FIND_FL_SKIP_LIMIT		0x004
+#define PR_CONFIG_FIND_FL_SKIP_DYNDIR		0x008
+
 config_rec *find_config_next(config_rec *, config_rec *, int,
   const char *, int);
+config_rec *find_config_next2(config_rec *, config_rec *, int,
+  const char *, int, unsigned long);
 config_rec *find_config(xaset_t *, int, const char *, int);
+config_rec *find_config2(xaset_t *, int, const char *, int, unsigned long);
 void find_config_set_top(config_rec *);
+
 int remove_config(xaset_t *, const char *, int);
 
+#define PR_CONFIG_FL_INSERT_HEAD	0x001
+config_rec *pr_config_add_set(xaset_t **, const char *, int);
+config_rec *pr_config_add(server_rec *, const char *, int);
+
 /* Returns the assigned ID for the provided directive name, or zero
  * if no ID mapping was found.
  */
